@@ -11,7 +11,7 @@
 '''
 
 import json
-from .config import data_path, cal_config
+
 from joblib import Parallel, delayed
 import copy
 import os
@@ -20,13 +20,18 @@ import pandas as pd
 import numpy as np
 import time
 import h5py
-import numba as nb
-from pathlib import Path
-from .tools import bisect_left, bisect_right, to_intdate
+import logging
 from pyarrow.dataset import field as ffield
 from pyarrow import dataset as ds
-
+from pathlib import Path
+from .tools import bisect_left, bisect_right, to_intdate, read_version_config
+from .config import data_path, cal_config
 markets = ["zz500", "hs300", "zz800", "zz1000", "all"]
+logger = logging.getLogger("logger")
+stream_handler = logging.StreamHandler()
+stream_handler.setLevel(logging.WARNING)
+formatter = logging.Formatter("%(asctime)s %(name)s %(levelname)s %(message)s")
+logger.addHandler(stream_handler)
 
 
 class BaseProvider:
@@ -68,7 +73,7 @@ class BaseProvider:
         """
         return FeatureD.feature_info(field, freq)
 
-    def features(self, instruments, fields, start_time=None, end_time=None, freq="Tday", return_mat=False):
+    def features(self, fields, start_time=None, end_time=None, freq="Tday", return_mat=False, instruments="all"):
         """
         :param instruments: instruments list or market (ei. "zz500", "hs300", "zz800", "zz1000", "all"
         :param fields: features str or list
@@ -85,18 +90,23 @@ class BaseProvider:
         fields = list(fields)  # In case of tuple.
         return FeatureD.features(instruments, fields, start_time, end_time, freq, return_mat)
 
-    def features_tday(self, instruments, fields, start_time=None, end_time=None, return_mat=False):
+    def features_mmat(self, fields, start_time=None, end_time=None, freq="Tday", instruments="all"):
         """
         :param instruments: instruments list or market (ei. "zz500", "hs300", "zz800", "zz1000", "all"
-        :param fields:  features str or list
+        :param fields: features str or list
         :param start_time:
         :param end_time:
-        :param return_mat:
-        :return:
+        :param freq: includdes Aday for natural day; Tday for trading day; report_period for financial statement
+        :return: DataFrames that may have different shapes
         """
-        return self.features(instruments, fields, start_time, end_time, "Tday", return_mat)
+        if isinstance(fields, str):
+            fields = [fields, ]
+        if freq == "report_period":
+            fields += ["ann_date", ]
+        fields = list(fields)  # In case of tuple.
+        return FeatureD.features_mmat(instruments, fields, start_time, end_time, freq)
 
-    def features_aday(self, instruments, fields, start_time=None, end_time=None, return_mat=False):
+    def features_tday(self, fields, start_time=None, end_time=None, return_mat=False, instruments="all"):
         """
         :param instruments: instruments list or market (ei. "zz500", "hs300", "zz800", "zz1000", "all"
         :param fields:  features str or list
@@ -105,9 +115,9 @@ class BaseProvider:
         :param return_mat:
         :return:
         """
-        return self.features(instruments, fields, start_time, end_time, "Aday", return_mat)
+        return self.features(fields, start_time, end_time, "Tday", return_mat, instruments)
 
-    def features_rpday(self, instruments, fields, start_time=None, end_time=None, return_mat=False):
+    def features_aday(self, fields, start_time=None, end_time=None, return_mat=False, instruments="all"):
         """
         :param instruments: instruments list or market (ei. "zz500", "hs300", "zz800", "zz1000", "all"
         :param fields:  features str or list
@@ -116,13 +126,24 @@ class BaseProvider:
         :param return_mat:
         :return:
         """
-        return self.features(instruments, fields, start_time, end_time, "report_period", return_mat)
+        return self.features(fields, start_time, end_time, "Aday", return_mat, instruments)
+
+    def features_rpday(self, fields, start_time=None, end_time=None, return_mat=False, instruments="all"):
+        """
+        :param instruments: instruments list or market (ei. "zz500", "hs300", "zz800", "zz1000", "all"
+        :param fields:  features str or list
+        :param start_time:
+        :param end_time:
+        :param return_mat:
+        :return:
+        """
+        return self.features(fields, start_time, end_time, "report_period", return_mat, instruments)
 
     def list_datasets(self):
         """
         :return:  list of dataset name
         """
-        return datasets_ls
+        return list_datasets()
 
     def dataset_info(self, dataset):
         """
@@ -164,13 +185,19 @@ class BaseProvider:
         fields = list(fields)
         return DatasetD.datasets_fl(filt, fields, dataset, **kwargs)
 
+    def version_info(self):
+        return read_version_config()
+
 
 class CalendarProvider:
     """Calendar provider base class: Provide calendar data."""
 
     def __init__(self):
         self.cal_cache = {}
-        self._uri_cal = os.path.join(data_path, "calendars", "{}.txt")
+
+    @property
+    def _uri_cal(self):
+        return os.path.join(data_path, "calendars", "{}.txt")
 
     def cache_cal(self, freq):
         _calendar = np.array(self.load_calendar(freq))
@@ -329,7 +356,7 @@ class BaseDataProvider:
                 instruments_d = Inst.list_instruments(instruments=instruments, start_time=start_time,
                                                       end_time=end_time, freq=freq, as_list=False)
             else:
-                raise ValueError("market not found")
+                instruments_d = [instruments, ]
 
         elif isinstance(instruments, (list, tuple, pd.Index, np.ndarray)):
             # list or tuple of a group of instruments
@@ -396,6 +423,17 @@ class FeatureProvider(BaseDataProvider):
                                       return_mat)
         return data
 
+    def features_mmat(self, instruments, fields, start_time=None, end_time=None, freq="Tday"):
+        instruments_d = self.get_instruments_d(instruments, start_time, end_time, freq)
+        column_names = self.get_column_names(fields)
+        cal = Cal.calendar(start_time, end_time, freq)
+        if len(cal) == 0:
+            return pd.DataFrame(columns=column_names)
+        start_time = cal[0]
+        end_time = cal[-1]
+        data = self.feature_processor_mmat(self._uri_data, instruments_d, column_names, start_time, end_time, freq)
+        return data
+
     def feature_info(self, field, freq):
         config_path = self._uri_cfg.format(freq, field)
         return FeatureD.read_config(config_path)
@@ -410,15 +448,15 @@ class FeatureProvider(BaseDataProvider):
         # One process for one task, so that the memory will be freed quicker.
         if isinstance(instruments_d, dict):
             instruments_d = list(instruments_d.keys())
-        start_index, end_index = Cal.locate_index(start_time, end_time, freq)
+        instruments_d = np.array(instruments_d)
         dtname = cal_config[freq]["col_name"]
-        timeindex, mask, codes = read_mask(fpath.format(freq, dtname), instruments_d, start_index, end_index,
-                                           return_mat)
-        if mask is None:
+        normalize_column_names = FeatureD.normalize_cache_fields(column_names)
+        stk_univ, time_idx, inst_mask, time_mask, mmp_shape = read_mask(fpath.format(freq, dtname), instruments_d,
+                                                                        start_time, end_time, freq)
+        if not time_mask.any():
             return pd.DataFrame(columns=column_names)
 
         workers = min(kernels, len(instruments_d))
-        normalize_column_names = FeatureD.normalize_cache_fields(column_names)
         start = time.perf_counter()
 
         field_l = []
@@ -427,7 +465,7 @@ class FeatureProvider(BaseDataProvider):
             field_l.append(field)
             tasks_l.append(
                 delayed(read_mmap)(
-                    fpath.format(freq, field), 'data', mask
+                    fpath.format(freq, field), "data", inst_mask, time_mask, mmp_shape
                 )
             )
 
@@ -439,14 +477,55 @@ class FeatureProvider(BaseDataProvider):
         )
         if return_mat:
             data = {
-                k: v.reshape((len(instruments_d), -1)).T for k, v in data.items()
+                k: v.T for k, v in data.items() if v is not None
             }
-            data["stockcode"] = codes
-            data[dtname] = Cal.calendar(start_time, end_time, freq)
+            data["stockcode"] = stk_univ[inst_mask]
+            data[dtname] = time_idx[time_mask]
         else:
-            data["stockcode"] = codes
-            data[dtname] = timeindex
-            data = pd.DataFrame(data).set_index(["stockcode", dtname]).dropna(how="all")
+            data = {
+                k: v.flatten() for k, v in data.items() if v is not None
+            }
+            multiindex = pd.MultiIndex.from_product([stk_univ[inst_mask], time_idx[time_mask]],
+                                                    names=["stockcode", cal_config[freq]["col_name"]])
+            data = pd.DataFrame(data, index=multiindex).reset_index()
+        print(time.perf_counter() - start)
+        return data
+
+    @staticmethod
+    def feature_processor_mmat(fpath, instruments_d, column_names, start_time, end_time, freq):
+        """
+        Load and process the data, return the data set.
+        - default using multi-kernel method.
+
+        """
+        # One process for one task, so that the memory will be freed quicker.
+        if isinstance(instruments_d, dict):
+            instruments_d = list(instruments_d.keys())
+        instruments_d = np.array(instruments_d)
+        normalize_column_names = FeatureD.normalize_cache_fields(column_names)
+
+        workers = min(kernels, len(instruments_d))
+        start = time.perf_counter()
+
+        field_l = []
+        tasks_l = []
+        for field in normalize_column_names:
+            field_l.append(field)
+            tasks_l.append(
+                delayed(read_mmat)(
+                    fpath.format(freq, field), instruments_d, start_time, end_time, freq
+                )
+            )
+
+        data = dict(
+            zip(
+                field_l,
+                Parallel(n_jobs=workers, backend="threading")(tasks_l),
+            )
+        )
+        data = {
+            k: v for k, v in data.items() if v is not None
+        }
         print(time.perf_counter() - start)
         return data
 
@@ -506,7 +585,7 @@ class DatasetProvider(BaseDataProvider):
         primary_keys = [code_flt, time_flt]  # 插入primary key
         if is_annrepo:
             primary_keys += ["report_period"]
-        normalize_column_names = normalize_column_names = DatasetD.normalize_cache_fields(primary_keys + normalize_column_names)
+        normalize_column_names = DatasetD.normalize_cache_fields(primary_keys + normalize_column_names)
         code_flt = ffield(code_flt) if code_flt is not None else None
         time_flt = ffield(time_flt) if time_flt is not None else None
         dset = ds.dataset(fpath, format=format_)  # 还差不规则特征
@@ -582,8 +661,28 @@ def list_datasets():
     return datasets
 
 
-def read_mmap(pth, key, mask=None):
+def read_mmat(pth, ori_instruments, start_time, end_time, freq):
     """numpy 数据结构提取, mmap加速"""
+    name = Path(pth).name.split(".")[0]
+    if not os.path.exists(pth):
+        logger.warning(f"WARN feature {name} not found")
+        return pd.DataFrame()
+    stk_univ, time_idx, inst_mask, time_mask, shape = read_mask(pth, ori_instruments, start_time, end_time, freq)
+    arr = read_mmap(pth, "data", inst_mask, time_mask, shape)
+    if arr is None:
+        return pd.DataFrame()
+    opt = pd.DataFrame(arr.T, index=time_idx[time_mask], columns=stk_univ[inst_mask])
+    opt.index.names = [cal_config[freq]["col_name"]]
+    opt.columns.names = ["stockcode"]
+    return opt
+
+
+def read_mmap(pth, key, inst_mask, time_mask, mmp_shape):
+    """numpy 数据结构提取, mmap加速"""
+    name = Path(pth).name.split(".")[1]
+    if not os.path.exists(pth):
+        logger.warning(f"WARN feature {name} not found")
+        return None
     with h5py.File(pth, 'r') as f:
         ds = f[key]
         offset = ds.id.get_offset()
@@ -593,55 +692,34 @@ def read_mmap(pth, key, mask=None):
         dtype = ds.dtype
         shape = ds.shape
     arr = np.memmap(pth, mode='r', shape=shape, offset=offset, dtype=dtype)
-    if not mask is None:
-        return arr[mask]
-    return arr
+    try:
+        arr = arr.reshape(mmp_shape)
+        return arr[:, time_mask][inst_mask, :]
+    except:
+        name = Path(pth).name.split(".")[1]
+        logger.warning(f"mask mismatch, feature {name} has different shape, please load it seperately")
+        return None
+
+
+def constru_index(pth, freq):
+    start_time, end_time = read_index(pth)
+    stk_univ = sorted(list(Inst.list_instruments("all", start_time, end_time, freq).keys()))
+    time_idx = Cal.calendar(start_time, end_time, freq=freq)
+    return np.array(stk_univ), time_idx, (len(stk_univ), len(time_idx))
 
 
 def read_index(pth):
     with h5py.File(pth, 'r') as f:
-        instlist = f['instlist'][:].astype(str)  # inst code list
-        instloc = f['instloc'][:]
-        rsilist = f['rsilist'][:]
-        reilist = f['reilist'][:]
-        cidx = f['data'].shape[0]
-    return instlist, instloc, rsilist, reilist, cidx
+        start_time = f["data"].attrs["start_time"]
+        end_time = f["data"].attrs["end_time"]
+    return start_time, end_time
 
 
-def read_mask(path, ori_instruments, start_index, end_index, mat_mode=False):
-    instlist, instloc, rsilist, reilist, cidx = read_index(path)
-    timeindex = read_mmap(path, "data")
-    mask, codes = None, []
-    codes, infoloc = get_availables(ori_instruments, instlist)
-    if len(infoloc) > 0:
-        mask, amounts = gen_idxs(cidx, infoloc, instloc, rsilist, reilist, start_index, end_index)
-        if not mat_mode:  # 返回矩阵形式
-            codes = np.repeat(codes, amounts)
-            timeindex = timeindex[mask]
-    return timeindex, mask, codes
-
-
-def get_availables(instruments, availables):
-    series = pd.Series(np.arange(len(availables)), index=availables).reindex(instruments).dropna()
-    return series.index.values, series.values.astype(int)
-
-
-@nb.jit(nopython=True)
-def gen_idxs(cidx, infoloc, instloc, rsilist, reilist, start_index, end_index):
-    mask = np.zeros(cidx, dtype=np.bool_)
-    amounts = []
-    for i in infoloc:
-        if (end_index < rsilist[i]) | (start_index > reilist[i]) | (start_index > end_index):
-            amounts.append(0)
-            continue
-        instsiloc = instloc[i]
-        tsi = max(start_index - rsilist[i], 0)
-        tei = min(end_index - rsilist[i], reilist[i] - rsilist[i])
-        h5si, h5ei = tsi + instsiloc, tei + instsiloc
-        amounts.append(h5ei - h5si + 1)
-        mask[h5si:h5ei + 1] = True
-        assert amounts[-1] > 0
-    return mask, amounts
+def read_mask(path, ori_instruments, start_time, end_time, freq):
+    stk_univ, time_idx, shape = constru_index(path, freq)
+    inst_mask = np.in1d(stk_univ, ori_instruments)
+    time_mask = (time_idx >= start_time) & (time_idx <= end_time)
+    return stk_univ, time_idx, inst_mask, time_mask, shape
 
 
 def parse_field(field):
@@ -650,9 +728,6 @@ def parse_field(field):
     return re.sub(r"\$(\w+)", r'ffield("\1")', field)
 
 
-features_ls = {**{a: "Aday" for a in list_features("Aday")}, **{t: "Tday" for t in list_features("Tday")},
-               **{rp: "report_period" for rp in list_features("report_period")}}
-datasets_ls = list_datasets()
 DatasetD = DatasetProvider()
 FeatureD = FeatureProvider()
 Cal = CalendarProvider()
